@@ -4,11 +4,11 @@ pick! {
   if #[cfg(target_feature="avx2")] {
     #[derive(Default, Clone, Copy, PartialEq, Eq)]
     #[repr(C, align(32))]
-    pub struct u32x8 { avx2: m256i }
+    pub struct u32x8 { pub(crate) avx2: m256i }
   } else {
     #[derive(Default, Clone, Copy, PartialEq, Eq)]
     #[repr(C, align(32))]
-    pub struct u32x8 { a : u32x4, b : u32x4 }
+    pub struct u32x8 { pub(crate) a : u32x4, pub(crate) b : u32x4 }
   }
 }
 
@@ -125,6 +125,35 @@ impl BitXor for u32x8 {
   }
 }
 
+impl From<u16x8> for u32x8 {
+  /// widens and zero extends to u32x8
+  #[inline]
+  #[must_use]
+  fn from(v: u16x8) -> Self {
+    pick! {
+      if #[cfg(target_feature="avx2")] {
+        Self { avx2:convert_to_i32_m256i_from_u16_m128i(v.sse) }
+      } else if #[cfg(target_feature="sse2")] {
+        Self {
+          a: u32x4 { sse: shr_imm_u32_m128i::<16>( unpack_low_i16_m128i(v.sse, v.sse)) },
+          b: u32x4 { sse: shr_imm_u32_m128i::<16>( unpack_high_i16_m128i(v.sse, v.sse)) },
+        }
+      } else {
+        u32x8::new([
+          u32::from(v.as_array_ref()[0]),
+          u32::from(v.as_array_ref()[1]),
+          u32::from(v.as_array_ref()[2]),
+          u32::from(v.as_array_ref()[3]),
+          u32::from(v.as_array_ref()[4]),
+          u32::from(v.as_array_ref()[5]),
+          u32::from(v.as_array_ref()[6]),
+          u32::from(v.as_array_ref()[7]),
+        ])
+      }
+    }
+  }
+}
+
 macro_rules! impl_shl_t_for_u32x8 {
   ($($shift_type:ty),+ $(,)?) => {
     $(impl Shl<$shift_type> for u32x8 {
@@ -176,6 +205,58 @@ macro_rules! impl_shr_t_for_u32x8 {
 
 impl_shr_t_for_u32x8!(i8, u8, i16, u16, i32, u32, i64, u64, i128, u128);
 
+/// Shifts lanes by the corresponding lane.
+///
+/// Bitwise shift-right; yields `self >> mask(rhs)`, where mask removes any
+/// high-order bits of `rhs` that would cause the shift to exceed the bitwidth of
+/// the type. (same as `wrapping_shr`)
+impl Shr<u32x8> for u32x8 {
+  type Output = Self;
+
+  #[inline]
+  #[must_use]
+  fn shr(self, rhs: u32x8) -> Self::Output {
+    pick! {
+      if #[cfg(target_feature="avx2")] {
+        // ensure same behavior as scalar wrapping_shr
+        let shift_by = bitand_m256i(rhs.avx2, set_splat_i32_m256i(31));
+        Self { avx2: shr_each_u32_m256i(self.avx2, shift_by ) }
+      } else {
+        Self {
+          a : self.a.shr(rhs.a),
+          b : self.b.shr(rhs.b),
+        }
+      }
+    }
+  }
+}
+
+/// Shifts lanes by the corresponding lane.
+///
+/// Bitwise shift-left; yields `self << mask(rhs)`, where mask removes any
+/// high-order bits of `rhs` that would cause the shift to exceed the bitwidth of
+/// the type. (same as `wrapping_shl`)
+impl Shl<u32x8> for u32x8 {
+  type Output = Self;
+
+  #[inline]
+  #[must_use]
+  fn shl(self, rhs: u32x8) -> Self::Output {
+    pick! {
+      if #[cfg(target_feature="avx2")] {
+        // ensure same behavior as scalar wrapping_shl
+        let shift_by = bitand_m256i(rhs.avx2, set_splat_i32_m256i(31));
+        Self { avx2: shl_each_u32_m256i(self.avx2, shift_by) }
+      } else {
+        Self {
+          a : self.a.shl(rhs.a),
+          b : self.b.shl(rhs.b),
+        }
+      }
+    }
+  }
+}
+
 impl u32x8 {
   #[inline]
   #[must_use]
@@ -201,7 +282,9 @@ impl u32x8 {
   pub fn cmp_gt(self, rhs: Self) -> Self {
     pick! {
       if #[cfg(target_feature="avx2")] {
-        Self { avx2: cmp_gt_mask_i32_m256i(self.avx2, rhs.avx2 ) }
+        // no unsigned gt than so inverting the high bit will get the correct result
+        let highbit = u32x8::splat(1 << 31);
+        Self { avx2: cmp_gt_mask_i32_m256i((self ^ highbit).avx2, (rhs ^ highbit).avx2 ) }
       } else {
         Self {
           a : self.a.cmp_gt(rhs.a),
@@ -210,20 +293,14 @@ impl u32x8 {
       }
     }
   }
+
   #[inline]
   #[must_use]
   pub fn cmp_lt(self, rhs: Self) -> Self {
-    pick! {
-      if #[cfg(target_feature="avx2")] {
-        Self { avx2: !cmp_gt_mask_i32_m256i(self.avx2, rhs.avx2) ^ cmp_eq_mask_i32_m256i(self.avx2,rhs.avx2) }
-      } else {
-        Self {
-          a : self.a.cmp_lt(rhs.a),
-          b : self.b.cmp_lt(rhs.b),
-        }
-      }
-    }
+    // lt is just gt the other way around
+    rhs.cmp_gt(self)
   }
+
   #[inline]
   #[must_use]
   pub fn blend(self, t: Self, f: Self) -> Self {
@@ -244,7 +321,7 @@ impl u32x8 {
   pub fn max(self, rhs: Self) -> Self {
     pick! {
       if #[cfg(target_feature="avx2")] {
-        Self { avx2: max_i32_m256i(self.avx2, rhs.avx2 ) }
+        Self { avx2: max_u32_m256i(self.avx2, rhs.avx2 ) }
       } else {
         Self {
           a : self.a.max(rhs.a),
@@ -258,7 +335,7 @@ impl u32x8 {
   pub fn min(self, rhs: Self) -> Self {
     pick! {
       if #[cfg(target_feature="avx2")] {
-        Self { avx2: max_i32_m256i(self.avx2, rhs.avx2 ) }
+        Self { avx2: min_u32_m256i(self.avx2, rhs.avx2 ) }
       } else {
         Self {
           a : self.a.min(rhs.a),
@@ -268,6 +345,36 @@ impl u32x8 {
     }
   }
 
+  #[inline]
+  #[must_use]
+  pub fn any(self) -> bool {
+    pick! {
+      if #[cfg(target_feature="avx2")] {
+        ((move_mask_i8_m256i(self.avx2) as u32) & 0b10001000100010001000100010001000) != 0
+      } else {
+        (self.a | self.b).any()
+      }
+    }
+  }
+
+  #[inline]
+  #[must_use]
+  pub fn all(self) -> bool {
+    pick! {
+      if #[cfg(target_feature="avx2")] {
+        ((move_mask_i8_m256i(self.avx2) as u32) & 0b10001000100010001000100010001000) == 0b10001000100010001000100010001000
+      } else {
+        (self.a & self.b).all()
+      }
+    }
+  }
+
+  #[inline]
+  #[must_use]
+  pub fn none(self) -> bool {
+    !self.any()
+  }
+
   #[inline]
   pub fn to_array(self) -> [u32; 8] {
     cast(self)
